# -*- shell-script -*-
#
# bash-ini: functions for handling ini files in bash
# Copyright (C) 2015-2018 Heinz Deinhart
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# bash-ini version 0.4

# standard error return codes:
#   1 .. local error
#   2 .. illegal number of arguments in function call
#   3 .. illegal retval-name
#   4 .. illegal db-name
#   5 .. db named db-name not loaded
#   6 .. section not found
#   7 .. key not found

# _bi_validate_dbname <caller-name> <db-name>
# errors: 1 (invalid db-name), 2
function _bi_validate_dbname() {
    (( $# == 1 )) || {
	bb_log "_bi_validate_dbname called with $# args instead of 1"
	return 2
    }

    [[ "$1" =~ ^[[:alnum:]_]+$ ]]
}

# _bi_validate_retvn <caller-name> <retvn>
# errors: 1 (invalid retvn), 2
function _bi_validate_retvn() {
    (( $# == 1 )) || {
	bb_log "_bi_validate_retvn called with $# args instead of 1"
	return 2
    }

    [[ "$1" =~ ^[[:alpha:]][[:alnum:]_]*$ ]]
}

# bi_db_load <db-name> <ini-file>
# loads contents of ini-file
# db-name must match ^[[:alnum:]_]+$ and must be currently unset
# errors: 2, 4
#  11 .. db-name already set
#  12 .. ini-file not readable
#  13 .. parse error
function bi_db_load() {
    (( $# == 2 )) || {
	bb_log "bi_db_load called with $# args instead of 2"
	return 2
    }

    local dbname="$1"
    local inputfile="$2"

    _bi_validate_dbname "$dbname" || {
	bb_log_debug 1 "bi_db_load: abort: incorrect db-name: $dbname "
	return 4
    }

    if bi_db_is_loaded "$dbname"
    then
	bb_log_debug 1 \
	    "bi_db_load: abort: db named $dbname is already loaded/set"
	return 11
    fi

    [ -r "$inputfile" ] || {
	bb_log_debug 1 "bi_db_load: can not read file: $inputfile"
	return 12
    }

    declare -g -A "bi_db_$dbname"

    local -i linenr=0
    local section=

    local extglobsave=$(shopt -p extglob || true)
    shopt -s extglob

    while read -r line || [[ -n "$line" ]]
    do
	linenr=$((linenr + 1))

        # ignore empty lines and comments
	[[ "$line" =~ ^([[:space:]]*|[\;\#].*)$ ]] && {
	    continue
	}

	# sections
	[[ "$line" =~ ^\[([[:alnum:]_\-]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)\][[:space:]]*$ ]] && {
	    section=${BASH_REMATCH[1]}
	    eval "bi_db_$dbname[\"\$section§\"]=''"
	    continue
	}

	# values
	[[ "$line" =~ ^[[:space:]]*([[:alnum:]]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)[[:blank:]]*=[[:blank:]]*(.*)[[:blank:]]*$ ]] && {
	    local key=${BASH_REMATCH[1]}
	    # shellcheck disable=SC2034
	    local value=${BASH_REMATCH[3]}
	    eval "bi_db_$dbname[\"\$section§$key\"]=\"\$value\""
	    continue
	}

	bb_log_debug 1 "bi_db_load: parse error at line $linenr near \"$line\""
	eval "$extglobsave"
	return 13
    done < "$inputfile"

    eval "$extglobsave"
}

# bi_db_destroy <db-name>
# destroys loaded ini db
# errors 2, 4
function bi_db_destroy() {
    (( $# == 1 )) || {
	bb_log "bi_db_destroy called with $# args instead of 1"
	return 2
    }

    local dbname="$1"
    _bi_validate_dbname "$dbname" || return 4
    unset "bi_db_$dbname"
}

# bi_db_is_loaded <db-name>
# check if db is loaded
# errors 2, 4, 5
function bi_db_is_loaded() {
    (( $# == 1 )) || {
	bb_log "bi_db_is_loaded called with $# args instead of 1"
	return 2
    }

    local dbname=$1
    _bi_validate_dbname "$dbname" || return 4

    [[ -v "bi_db_${dbname}" ]] || return 5
}

# bi_key_exists <db-name> <key> [<section>]
# check if a key exists
# if no section is specified only global context is searched
# errors: 2, 4, 5, 6, 7
function bi_key_exists() {
    (( $# == 2 || $# == 3 )) || {
	bb_log "bi_key_exists needs 2 or 3 args but called with $#"
	return 2
    }

    local dbname="$1"
    local key="$2"
    local section="${3-}"

    if [[ -z "$section" ]]
    then
	bi_db_is_loaded "$dbname" || return $? # returns 4, 5
    else
	bi_section_exists "$dbname" "$section" || return $? # returns 4, 5, 6
    fi

    eval "[[ \"\${bi_db_${dbname}[\"\$section§\$key\"]+x}\" == \"x\" ]]" || \
	return 7
}

# bi_key_exists_with_default <db-name> <key> <section>
# like bi_key_exists but if section does exist and does not contain key it
# will search in global context
# errors: 2, 4, 5, 6, 7
function bi_key_exists_with_default() {
    (( $# == 3 )) || {
	bb_log "bi_key_exists_with_default needs 3 args but called with $#"
	return 2
    }

    local dbname="$1"
    local key="$2"
    local section="$3"

    bi_section_exists "$dbname" "$section" || return $? # returns 4, 5, 6
    bi_key_exists "$dbname" "$key" "$section" ||
	bi_key_exists "$dbname" "$key" # returns 4, 5, 6, 7
}

# bi_section_exists <db-name> <section>
# check if a section exists
# errors: 2, 4, 5, 6
function bi_section_exists() {
    (( $# == 2 )) || {
	bb_log "bi_section_exists needs 2 args but called with $#"
	return 2
    }

    local dbname="$1"
    local section="$2"

    bi_db_is_loaded "$dbname" || return $? # returns 4, 5

    eval "[[ \"\${bi_db_${dbname}[\"\$section§\"]+isset}\" == \"isset\" ]]" || \
	return 6
}

# bi_get_section_list <retval-name> <db-name>
# return a list of all sections as array into global array <retval-name>
# retval-name must match ^[[:alpha:]][[:alnum:]_]*$
# note: the order is undefined
# errors: 2, 3
function bi_get_section_list() {
    (( $# == 2 )) || {
	bb_log "bi_get_section_list needs 2 args but called with $#"
	return 2
    }

    local _retvn="$1"
    local _dbname="$2"

    _bi_validate_retvn "$_retvn" || return 3
    bi_db_is_loaded "$_dbname" || return $? # returns 4, 5

    local -a _keys
    eval "_keys=(\"\${!bi_db_${_dbname}[@]}\")"
    local -a _sections=()

    if (( ${#_keys[*]} > 0 )) # skip if file is empty
    then
	local _k
	for _k in "${_keys[@]}"
	do
	    [[ "$_k" =~ §$ ]] && _sections+=("${_k%§}")
	done
    fi

    if (( ${#_sections[@]} > 0 )) # any sections found?
    then # yes
	eval "${_retvn}=(\"\${_sections[@]}\")"
    else # no
	eval "${_retvn}=()"
    fi
}

# bi_get_value <retval-name> <db-name> <key> [ <section> ]
# return value for <key> in <db-name> in optional <section>
# note: if <section> is specified key must exist in it
# note: <retval-name> must not start with '_'
# errors: 1,2,3,4
function bi_get_value() {
    (( $# == 3 || $# == 4 )) || {
	bb_log "bi_get_value needs 3 or 4 args but called with $#"
	return 2
    }

    local _retvn="$1"
    local _dbname="$2"
    local _key="$3"
    local _section="${4-}"

    _bi_validate_retvn "$_retvn"

    bi_key_exists "$_dbname" "$_key" "$_section" || \
	return $? # returns 4, 5, 6, 7

    eval "${_retvn}=\"\${bi_db_$_dbname[\"\$_section§\$_key\"]}\""
}

# bi_get_value_with_default <variablename> <db-name> <_key> <_section>
# return value for <key> in <db-name> in <section>
# or try global if <section> exists but does not contain key
# note: <retval-name> must not start with '_'
function bi_get_value_with_default() {
    (( $# == 4 )) || {
	bb_log "bi_get_value_with_default needs 4 args but called with $#"
	return 2
    }

    local _retvn="$1"
    local _dbname="$2"
    local _key="$3"
    local _section="$4"

    _bi_validate_retvn "$_retvn"

    bi_get_value "$_retvn" "$_dbname" "$_key" "$_section" || {
	local _i="$?"
	case "$_i" in
	    7) # section was found, but key not
		bi_get_value "$_retvn" "$_dbname" "$_key";;
	    *) # section was not found
		return "$_i"
	esac
    }
}
