<!-- -*- mode: markdown; mode: flyspell -*- -->

# bash-ini library

A library to query [ini style](https://en.wikipedia.org/wiki/INI_file)
configuration files. Currently there is no support to write or change
files on disk.

## Security

This library currently makes heavy use of `eval`. This can not easily
change until bash 4.2 support is dropped. Until that it is probably a
good idea not using it in privileged scripts to parse ini files that
can be modified by untrusted users.

## Naming

To use the bash-ini, first the contents of an ini file must be loaded
into memory. There it is called database (db). This database is read
only and will not follow any changes on disk.

## ini file parsing

The ini file format is not standardized. bash-ini parses the following
file format:

1. Empty lines or comment lines (`^([[:space:]]*|[\;\#].*)$`) are ignored.
2. Any other line must be either a key value pair or a section definition
3. Section definitions must have the format `^\[([[:alnum:]_\-]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)\][[:space:]]*$`
4. Key value pairs must have the format `^[[:space:]]*([[:alnum:]]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)[[:blank:]]*=[[:blank:]]*(.*)[[:blank:]]*$`

## API Docu

### Error Codes

```bash
# standard error return codes:
#   1 .. local error
#   2 .. illegal number of arguments in function call
#   3 .. illegal retval-name
#   4 .. illegal db-name
#   5 .. db named db-name not loaded
#   6 .. section not found
#   7 .. key not found
```

### Functions

```bash
# bi_db_load <db-name> <ini-file>
# loads contents of ini-file
# db-name must match ^[[:alnum:]_]+$ and must be currently unset
# errors: 2, 4
#  11 .. db-name already set
#  12 .. ini-file not readable
#  13 .. parse error
```

```bash
# bi_db_destroy <db-name>
# destroys loaded ini db
# errors 2, 4
```

```bash
# bi_db_is_loaded <db-name>
# check if db is loaded
# errors 2, 4, 5
```

```bash
# bi_key_exists <db-name> <key> [<section>]
# check if a key exists
# if no section is specified only global context is searched
# errors: 2, 4, 5, 6, 7
```

```bash
# bi_key_exists_with_default <db-name> <key> <section>
# like bi_key_exists but if section does exist and does not contain key it
# will search in global context
# errors: 2, 4, 5, 6, 7
```

```bash
# bi_section_exists <db-name> <section>
# check if a section exists
# errors: 2, 4, 5, 6
```

```bash
# bi_get_section_list <retval-name> <db-name>
# return a list of all sections as array into global array <retval-name>
# retval-name must match ^[[:alpha:]][[:alnum:]_]*$
# note: the order is undefined
# errors: 2, 3
```

```bash
# bi_get_value <retval-name> <db-name> <key> [ <section> ]
# return value for <key> in <db-name> in optional <section>
# note: if <section> is specified key must exist in it
# note: <retval-name> must not start with '_'
# errors: 1,2,3,4
```

```bash
# bi_get_value_with_default <variablename> <db-name> <_key> <_section>
# return value for <key> in <db-name> in <section>
# or try global if <section> exists but does not contain key
# note: <retval-name> must not start with '_'
```

## Example

The following ini file

```ini
;i this is a comment
mainval1=take this
main val 2=this too
x = y

[sec1]
sec1val1=huhu  
sec1val2=haha		
  
# this is a comment, too	
[sec 2]
sec2val1=hoho der xmas man is do
	sec2val2=tab
	one and one = 	three it is   	  		  
```
is accessed by the following demo code:
```bash
#!/bin/bash

source /usr/libexec/bash-base.bash || {
    echo "$0: fatal error: failed to source /usr/libexec/bash-base.bash" >&2
    exit 1
}

function main() {
    bi_db_load demo "$1" || bb_fatal "bi_db_load demo '$1' failed: $?"

    local -a sections
    local val

    bi_get_section_list sections demo
    echo "lisit of all sections: ${sections[*]}"

    bi_get_value val demo mainval1
    echo "key 'mainval1' in global context: '$val'"
    
    bi_get_value val demo sec2val1 "sec 2"
    echo "key 'sec2val1' in section 'sec 2': '$val'"

    bi_get_value val demo "one and one" "sec 2"
    echo "key 'one and one' in section 'sec 2': '$val'"

    echo "bi_demo done"

    bb_quit
}

main demo.ini
```
